;------------------------------------------------------------------------------
; Name: Timing macros for MASM.
; Version: 0.08
;  Author: Petri Partanen (aka. Petroizki)
;------------------------------------------------------------------------------
	
	PTIMER_RDTSC_LOOPS EQU 8   		; defines the loops for setup overhead timing
	PTIMER_LOOPS_POWER_OF_2 EQU 5   ; defines the power of the loop count (2^x), i just use this weird loopcount to get rid of divisions (1<=x<=31)
	                                ;  *BIG WARNING* the data section will grow by the multiply of 8, when loops are added
	PTIMER_WARMUP_LOOPS EQU 3       ; non-timed loops of the benchmark code (warmup phase)
	
	__PTIMER_LOOPS EQU (1 SHL PTIMER_LOOPS_POWER_OF_2)
	
	__ptimers_label EQU <>
	__ptimers_timer_count = 0
	
	;--------------------------------------------------------------------------
	; Desc: This is the macro that start single-pass timing.
	;--------------------------------------------------------------------------
	PTIMER_START_COUNT macro
		LOCAL lbl1, timer_label_local
		__ptimers_label EQU <timer_label_local>
		
		;; nag the user, if he/she is trying to nest the timers
		IF __ptimers_timer_count NE 0
			echo *************************************************************
			echo * ptimers forced ERROR.
			echo * Reason: Nesting of the timers is not allowed.
			echo *          Use the end macro, before starting a new one.
			echo *************************************************************
			
			.ERR
			exitm
		ELSE
			__ptimers_timer_count = 1
		ENDIF
		
		;; init the timers
		push OFFSET timer_label_local
		jmp __ptimers_init
		
		ALIGN 16
timer_label_local:
		push dword ptr [__PTIMER_SAFE_FLAGS]
		xor eax, eax
		popfd
		cpuid
		rdtsc
		mov dword ptr [__PTIMER_TIMING], eax
		mov dword ptr [__PTIMER_TIMING + 4], edx
		;; restore registers and flags, as they were before the macro call
		mov eax, dword ptr [__PTIMER_SAFE]
		mov ecx, dword ptr [__PTIMER_SAFE + 4]
		mov edx, dword ptr [__PTIMER_SAFE + 8]
		mov ebx, dword ptr [__PTIMER_SAFE + 12]
	endm
	
	;--------------------------------------------------------------------------
	; Desc: This is the macro that starts the loop timing.
	;--------------------------------------------------------------------------
	PTIMER_START_LOOP_COUNT macro
		mov dword ptr [__PTIMER_COUNT], (__PTIMER_LOOPS + PTIMER_WARMUP_LOOPS)
		
		PTIMER_START_COUNT
		
		__ptimers_timer_count = 2
	endm
	
	;--------------------------------------------------------------------------
	; Desc: This macro end single pass count
	;--------------------------------------------------------------------------
	PTIMER_END_COUNT macro
		IF __ptimers_timer_count NE 1
			echo *************************************************************
			echo * ptimers forced ERROR.
			echo * Reason: End timer macro used without correct start macro.
			echo *************************************************************
			
			.ERR
			exitm
		ENDIF
	
		__ptimers_timer_count = 0
		
		;; saves the state of the registers
		;; and gets the current time
		mov dword ptr [__PTIMER_SAFE + 16], eax
		mov dword ptr [__PTIMER_SAFE + 20], ecx
		mov dword ptr [__PTIMER_SAFE + 24], edx
		mov dword ptr [__PTIMER_SAFE + 28], ebx
		mov eax, 0
		cpuid
		rdtsc
		
		pushfd
		sub eax, dword ptr [__PTIMER_TIMING]
		sbb edx, dword ptr [__PTIMER_TIMING + 4]
		sub eax, dword ptr [__PTIMER_INIT_TIME]
		sbb edx, dword ptr [__PTIMER_INIT_TIME + 4]
		pop dword ptr [__PTIMER_SAFE_FLAGS + 4]
		
		mov dword ptr [qwPTIMER_RESULT], eax
		mov dword ptr [qwPTIMER_RESULT + 4], edx
		
		call __ptimers_results_end
	endm
	
	;--------------------------------------------------------------------------
	;--------------------------------------------------------------------------
	PTIMER_END_LOOP_COUNT macro
		IF __ptimers_timer_count NE 2
			echo *************************************************************
			echo * ptimers forced ERROR.
			echo * Reason: End timer macro used without correct start macro.
			echo *************************************************************
			
			.ERR
			exitm
		ENDIF
		
		__ptimers_timer_count = 0
		
		;; saves the state of the registers
		;; and gets the current time
		mov dword ptr [__PTIMER_SAFE + 16], eax
		mov dword ptr [__PTIMER_SAFE + 20], ecx
		mov dword ptr [__PTIMER_SAFE + 24], edx
		mov dword ptr [__PTIMER_SAFE + 28], ebx
		mov eax, 0
		cpuid
		rdtsc
		
		pushfd
		mov ecx, dword ptr [__PTIMER_COUNT]
		sub ecx, 1
		mov dword ptr [__PTIMER_COUNT], ecx
		pop dword ptr [__PTIMER_SAFE_FLAGS + 4]
		
		;; jump back, if this is just a warmup phase
		cmp ecx, dword ptr __PTIMER_LOOPS
		jnb __ptimers_label
		
		sub eax, dword ptr [__PTIMER_TIMING]
		sbb edx, dword ptr [__PTIMER_TIMING + 4]
		sub eax, dword ptr [__PTIMER_INIT_TIME]
		sbb edx, dword ptr [__PTIMER_INIT_TIME + 4]
		
		;; save the benchmarked time to the clock count table
		mov dword ptr [__PTIMER_TABLE + ecx*8], eax
		mov dword ptr [__PTIMER_TABLE + ecx*8 + 4], edx
		
		;; see if we have more loops left
		test ecx, ecx
		jnz __ptimers_label
		
		;; call final preparations
		call __ptimers_results
	endm
.data?
	ALIGN 16
	__PTIMER_INIT_TIME dq ?
	__PTIMER_TIMING dq ?
	__PTIMER_SAFE dd 8 dup (?)
	__PTIMER_SAFE_FLAGS dd 2 dup (?)
	__PTIMER_COUNT dd ?
	
	ALIGN 16
	qwPTIMER_RESULT dq ?
	__PTIMER_TABLE dq __PTIMER_LOOPS dup (?)
	__PTIMER_DIFFERENCE_TABLE dq __PTIMER_LOOPS dup (?)
.code
;==============================================================================
; Desc: Initialization of the timers.
;==============================================================================
	ALIGN 16
__ptimers_init_loop:
	push dword ptr [__PTIMER_SAFE_FLAGS]
	xor eax, eax
	popfd
	cpuid
	rdtsc
	mov dword ptr [__PTIMER_TIMING], eax
	mov dword ptr [__PTIMER_TIMING + 4], edx
	mov eax, dword ptr [__PTIMER_SAFE]
	mov ecx, dword ptr [__PTIMER_SAFE + 4]
	mov edx, dword ptr [__PTIMER_SAFE + 8]
	mov ebx, dword ptr [__PTIMER_SAFE + 12]
	; <- this would be the operation here
	mov dword ptr [__PTIMER_SAFE + 16], eax
	mov dword ptr [__PTIMER_SAFE + 20], ecx
	mov dword ptr [__PTIMER_SAFE + 24], edx
	mov dword ptr [__PTIMER_SAFE + 28], ebx
	mov eax, 0
	cpuid
	rdtsc
	
	pushfd
	; subtract the init time
	sub eax, dword ptr [__PTIMER_TIMING]
	sbb edx, dword ptr [__PTIMER_TIMING + 4]
	pop dword ptr [__PTIMER_SAFE_FLAGS + 4]
	
	mov ebx, edx
	cmp eax, ebp
	sbb ebx, esi
	jnb @F
	
	mov ebp, eax
	mov esi, edx
	
@@:	sub edi, 1
	jnz __ptimers_init_loop
	
	mov dword ptr [__PTIMER_INIT_TIME], ebp
	mov dword ptr [__PTIMER_INIT_TIME + 4], esi
	
	retn
	
	ALIGN 16
__ptimers_init:
	push edi
	mov edi, (OFFSET __PTIMER_SAFE)
	push ebp
	push esi

	; Preserve registers and flags.
	pushfd
	mov dword ptr [edi], eax
	mov dword ptr [edi + 4], ecx
	mov dword ptr [edi + 8], edx
	mov dword ptr [edi + 12], ebx
	pop dword ptr [__PTIMER_SAFE_FLAGS]
	
	mov edi, PTIMER_RDTSC_LOOPS
	
	or ebp, -1
	or esi, -1
	
	call __ptimers_init_loop
	
	pop esi
	pop ebp
	pop edi
	
	retn
	
;==============================================================================
; Desc: This stuff calculates the average clock count, ignoring the obviously
;        wrong ones. It also restores the state of registers and flags, like
;        they were after the timed code.
;==============================================================================
	ALIGN 16
__ptimers_results:
	push ebp
	push edi
	push esi

	mov ecx, __PTIMER_LOOPS
	mov ebp, (OFFSET __PTIMER_TABLE) - 8
	
	; calculate average of all timings (average = (x1+x2...+x3)/n)
	xor eax, eax
	xor edx, edx
@@:		add eax, dword ptr [ebp + ecx*8]
		adc edx, dword ptr [ebp + ecx*8 + 4]
		sub ecx, 1
	jnz @B
	; divide edx:eax with __PTIMER_LOOPS (= 2^PTIMER_LOOPS_POWER_OF_2)
	mov ebx, edx
	sar edx, PTIMER_LOOPS_POWER_OF_2
	shr eax, PTIMER_LOOPS_POWER_OF_2
	shl ebx, (32-PTIMER_LOOPS_POWER_OF_2)
	or eax, ebx
	; average in edx:eax
	
	; calculate standard deviation (s=sum(abs(x-average)))
	xor edi, edi
	mov ecx, __PTIMER_LOOPS
	push edi ; push 0
@@:		mov ebx, dword ptr [__PTIMER_TABLE + ecx*8 - 8]
		mov esi, dword ptr [__PTIMER_TABLE + ecx*8 - 4]
		sub ebx, eax
		sbb esi, edx
		
		; abs(esi:ebx)
		mov ebp, esi
		sar ebp, 31
		add esi, ebp
		add ebx, ebp
		xor esi, ebp
		xor ebx, ebp
		add esi, ebp
		
		add dword ptr [esp], ebx
		adc edi, esi
		
		mov dword ptr [__PTIMER_DIFFERENCE_TABLE + ecx*8 - 8], ebx
		mov dword ptr [__PTIMER_DIFFERENCE_TABLE + ecx*8 - 4], esi
	sub ecx, 1
	jnz @B
	pop edx
	mov ebx, edi
	sar edi, PTIMER_LOOPS_POWER_OF_2
	shr edx, PTIMER_LOOPS_POWER_OF_2
	shl ebx, (32-PTIMER_LOOPS_POWER_OF_2)
	or edx, ebx
	; standard deviation in edi:edx
	
	; calculate average again, this time ignoring all the timings that fall beyond standard deviation
	mov ecx, __PTIMER_LOOPS
	xor ebx, ebx
	add edx, 1 ; needed for carry flag check
	xor esi, esi
	xor eax, eax
__ptimers_continue:
		mov ebp, dword ptr [__PTIMER_DIFFERENCE_TABLE + ecx*8 - 4]
		cmp dword ptr [__PTIMER_DIFFERENCE_TABLE + ecx*8 - 8], edx
		sbb ebp, edi
		jnc @F
		
		; this is a good value, add it to the new average count
		add ebx, dword ptr [__PTIMER_TABLE + ecx*8 - 8]
		adc eax, dword ptr [__PTIMER_TABLE + ecx*8 - 4]
		add esi, 1
		
@@:		sub ecx, 1
		jnz __ptimers_continue
	xor edx, edx
	div esi
	mov dword ptr [qwPTIMER_RESULT + 4], eax
	mov eax, ebx
	div esi
	mov dword ptr [qwPTIMER_RESULT], eax
	; the final average
	
	pop esi
	pop edi
	pop ebp
	
__ptimers_results_end:
	mov ebx, (OFFSET __PTIMER_SAFE + 16)

	; restore flags, registers
	push dword ptr [__PTIMER_SAFE_FLAGS + 4]
	mov eax, dword ptr [ebx]
	mov ecx, dword ptr [ebx + 4]
	mov edx, dword ptr [ebx + 8]
	mov ebx, dword ptr [ebx + 12]
	popfd
	
	retn